/*
    This file is part of the oolaboola digital dj program

    Copyright (C) 1999  Eric S. Tiedemann

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <iostream>
#include <string>
#include <map>
#include <list>
#include <cctype>
#include <new>
#include <exception>
#include <cstdlib>

using namespace std;

class sexpr
{
  static sexpr read_refusing_EOF(istream& i);
  static char read_char_refusing_EOF(istream& i);
  static sexpr read_list(istream &i);
  static sexpr read_string(istream &i);
  static void consume_comment(istream &i);
  static sexpr read_atom(char c, istream &i);
  static string read_atom_string(char c, istream &i);
  static sexpr read_dot(istream &i);
  static sexpr read1(istream& i);
  static sexpr string2number(string s);
  static sexpr string2symbol(string s);
  static sexpr read(istream& i = cin);
  static void write_list_tail(sexpr &l, ostream &o);
  
  static bool terminatingp(int c);
  static map<string, sexpr *> symtab;

  // the data members
  string s; // can't go in the union cause it has a constructor
  union {
    struct {
      sexpr *car, *cdr;
    }p;
    long i;
    double f;
  }v;
    
  enum type {STRING, SYMBOL, PAIR, NIL, 
	     FIX, FLOAT,
	     CLOSE, DOT, EOFO} t;


  sexpr(sexpr::type t) : t(t) {}

public:

  sexpr() : t(NIL) {}		// default constructor

  sexpr(const sexpr &from)
    {
      if (from.t == PAIR)
	{
	  v.p.car = new sexpr(*from.v.p.car);
	  v.p.cdr = new sexpr(*from.v.p.cdr);
	  t = PAIR;
	}
      else
	{
	  t = from.t;
	  s = from.s;
	  v = from.v;
	}
    }

  ~sexpr()
    {
      if (t == PAIR)
	{
	  delete v.p.car;
	  delete v.p.cdr;
	}
    }

  sexpr & operator =(const sexpr &from)
    {
      if (t == PAIR)
	{
	  delete v.p.car;
	  delete v.p.cdr;
	}
      
      if (from.t == PAIR)
	{
	  v.p.car = new sexpr(*from.v.p.car);
	  v.p.cdr = new sexpr(*from.v.p.cdr);
	  t = PAIR;
	}
      else
	{
	  t = from.t;
	  s = from.s;
	  v = from.v;
	}

      return *this;
    }

  sexpr(double d) : t(FLOAT) { v.f = d; }

  sexpr(long l) : t(FIX) { v.i = l; }

  sexpr(sexpr *a, sexpr *b)
    : t(PAIR)
    {
      v.p.car = a;
      v.p.cdr = b;
    }

  sexpr(string s, bool symbolp = false)
    : s(s)
    {
      if (symbolp)
	t = SYMBOL;
      else
	t = STRING;
    }

  static ostream &write(sexpr &s, ostream& o = cout);

  class DomainError : public exception
  { public: char *what() { return "sexpr::DomainError"; } };

  class UnexpectedEOF : public exception
  { public: char *what() { return "sexpr::UnexpectedEOF"; } };

  class SyntaxError : public exception
  { public: char *what() { return "sexpr::SyntaxError"; } };

  sexpr
  &car()
      {
	if (t == PAIR)
	  return *v.p.car;
	else
	  throw DomainError();
      }
	
  sexpr
  &cdr()
      {
	if (t == PAIR)
	  return *v.p.cdr;
	else
	  throw DomainError();
      }
	
  bool 
  eof_objectp()
      {
	return t == EOFO;
      }

  bool nullp()
      {
	return t == NIL;
      }

  bool pairp()
      {
	return t == PAIR;
      }

  bool stringp()
      {
	return t == STRING;
      }

  bool floatp()
      {
	return t == FLOAT;
      }

  operator double()
    {
      if (t == FLOAT)
	return v.f;
      else
	return 0.0;
    }

  operator string()
    {
      return s;
    }

  static sexpr symbol(string s) { return sexpr(s, true); }

  static bool eq(const sexpr &, const sexpr &);

  friend istream& operator >>(istream&, sexpr&);
  friend ostream& operator <<(ostream&, sexpr&);
};
